@@ -18,12 +18,16 @@ using Mono.GetOptions;
 
 public class Stub {
 	
-	static Assembly assembly;
+	static string srcPath;
+	static Assembly[] assemblies;
 	
 	static bool nooverrides = true, delete = false, ignoremembers = false;
 	
 	static int additions = 0, deletions = 0;
 
+	static string name;
+	static XmlDocument slashdocs;
+	
 	private class Opts : Options {
 		[Option("The root {directory} of an assembly's documentation files.")]
 		public string path = null;
@@ -31,8 +35,8 @@ public class Stub {
 		[Option("When updating documentation, write the updated files to this {path}.")]
 		public string updateto = null;
 
-		[Option("The assembly to document.  Specify a {file} path or the name of a GAC'd assembly.")]
-		public string assembly = null;
+		[Option(-1, "The assembly to document.  Specify a {file} path or the name of a GAC'd assembly.")]
+		public string[] assembly = null;
 
 		[Option("Document only the {type name}d by this argument.")]
 		public string type = null;
@@ -48,6 +52,12 @@ public class Stub {
 
 		[Option("Don't update members.")]
 		public bool ignoremembers = false;
+
+		[Option("The {name} of the project this documentation is for.")]
+		public string name;
+
+		[Option("An XML documemntation {file} made by the /doc option of mcs/csc the contents of which will be imported.")]
+		public string importslashdoc;
 	}
 	
 	private static int n(object o) {
@@ -66,12 +76,15 @@ public class Stub {
 		nooverrides = !opts.overrides;
 		delete = opts.delete;
 		ignoremembers = opts.ignoremembers;
+		name = opts.name;
 
 		try {
 			// PARSE BASIC OPTIONS AND LOAD THE ASSEMBLY TO DOCUMENT
 			
 			if (opts.path == null)
 				throw new InvalidOperationException("The path option is required.");
+			
+			srcPath = opts.path;
 
 			if (n(opts.type) + n(opts.@namespace) > 1)
 				throw new InvalidOperationException("You cannot specify both 'type' and 'namespace'.");
@@ -79,18 +92,22 @@ public class Stub {
 			if (opts.assembly == null)
 				throw new InvalidOperationException("The assembly option is required.");
 				
-			try {
-				assembly = Assembly.LoadFile(opts.assembly);
-			} catch (System.IO.FileNotFoundException e) { }
-
-			if (assembly == null) {
-				try {
-					assembly = Assembly.LoadWithPartialName(opts.assembly);
-				} catch (Exception e) { }
-			}
+			assemblies = new Assembly [opts.assembly.Length];
+			for (int i = 0; i < opts.assembly.Length; i++)
+				assemblies [i] = LoadAssembly (opts.assembly [i]);
+				
+			// IMPORT FROM /DOC?
 			
-			if (assembly == null)
-				throw new InvalidOperationException("Assembly not found.");
+			if (opts.importslashdoc != null) {
+				try {
+					slashdocs = new XmlDocument();
+					slashdocs.Load(opts.importslashdoc);
+				} catch (Exception e) {
+					Console.Error.WriteLine("Could not load /doc file: " + e.Message);
+					Environment.ExitCode = 1;
+					return;
+				}
+			}
 			
 			// PERFORM THE UPDATES
 			
@@ -99,23 +116,45 @@ public class Stub {
 			else if (opts.@namespace != null)
 				DoUpdateNS(opts.@namespace, opts.path + "/" + opts.@namespace, opts.updateto == null ? opts.path + "/" + opts.@namespace : opts.updateto + "/" + opts.@namespace);
 			else
-				DoUpdateAssembly(opts.path, opts.updateto == null ? opts.path : opts.updateto);
+				DoUpdateAssemblies(opts.path, opts.updateto == null ? opts.path : opts.updateto);
 		
 		} catch (InvalidOperationException error) {
 			Console.Error.WriteLine(error.Message);
+			Environment.ExitCode = 1;
 			return;
 			
 		} catch (System.IO.IOException error) {
 			Console.Error.WriteLine(error.Message);
+			Environment.ExitCode = 1;
 			return;
 
 		} catch (Exception error) {
 			Console.Error.WriteLine(error);
+			Environment.ExitCode = 1;
 		}
 
 		Console.Error.WriteLine("Members Added: {0}, Members Deleted: {1}", additions, deletions);
 	}
 	
+	private static Assembly LoadAssembly (string name)
+	{
+		Assembly assembly = null;
+		try {
+			assembly = Assembly.LoadFile (name);
+		} catch (System.IO.FileNotFoundException e) { }
+
+		if (assembly == null) {
+			try {
+				assembly = Assembly.LoadWithPartialName (name);
+			} catch (Exception e) { }
+		}
+			
+		if (assembly == null)
+			throw new InvalidOperationException("Assembly " + name + " not found.");
+
+		return assembly;
+	}
+
 	private static void WriteXml(XmlElement element, System.IO.TextWriter output) {
 		XmlTextWriter writer = new XmlTextWriter(output);
 		writer.Formatting = Formatting.Indented;
@@ -131,12 +170,11 @@ public class Stub {
 		XmlElement index_root = index.CreateElement("Overview");
 		index.AppendChild(index_root);
 
-		XmlElement index_assembly = index.CreateElement("Assembly");
-		if (assembly == null)
+		if (assemblies.Length == 0)
 			throw new Exception ("No assembly");
-		index_assembly.InnerText = assembly.GetName().Name;
-		index_assembly.SetAttribute("Version", assembly.GetName().Version.ToString());
-		index_root.AppendChild(index_assembly);
+
+		XmlElement index_assemblies = index.CreateElement("Assemblies");
+		index_root.AppendChild(index_assemblies);
 
 		XmlElement index_remarks = index.CreateElement("Remarks");
 		index_remarks.InnerText = "To be added.";
@@ -149,8 +187,6 @@ public class Stub {
 		XmlElement index_types = index.CreateElement("Types");
 		index_root.AppendChild(index_types);
 		
-		MakeAttributes(index_root, assembly, true);
-		
 		return index;
 	}
 	
@@ -180,12 +216,17 @@ public class Stub {
 	}
 	
 	public static void DoUpdateType(string basepath, string typename, string dest) {
-		Type type = assembly.GetType(typename, false);
+		Type type = null;
+		foreach (Assembly assembly in assemblies) {
+			type = assembly.GetType(typename, false);
+			if (type != null)
+				break;
+		}
 		if (type == null) throw new InvalidOperationException("Type not found: " + typename);
+		if (type.Namespace == null) throw new InvalidOperationException("Types in the root namespace cannot be documented: " + typename);
 		string typename2 = type.FullName.Substring(type.Namespace == "" ? 0 : type.Namespace.Length+1);
 
 		XmlDocument basefile = new XmlDocument();
-		basefile.PreserveWhitespace = true;
 		string typefile = basepath + "/" + type.Namespace + "/" + typename2 + ".xml";
 		try {
 			basefile.Load(typefile);
@@ -205,10 +246,10 @@ public class Stub {
 	
 	public static void DoUpdateNS(string ns, string nspath, string outpath) {
 		Hashtable seenTypes = new Hashtable();
+		Assembly assembly = assemblies [0];
 		
 		foreach (System.IO.FileInfo file in new System.IO.DirectoryInfo(nspath).GetFiles("*.xml")) {
 			XmlDocument basefile = new XmlDocument();
-			basefile.PreserveWhitespace = true;
 			string typefile = nspath + "/" + file.Name;
 			try {
 				basefile.Load(typefile);
@@ -229,6 +270,7 @@ public class Stub {
 		
 		// Stub types not in the directory
 		foreach (Type type in assembly.GetTypes()) {
+			if (type.Namespace == null) continue;
 			if (type.Namespace != ns || seenTypes.ContainsKey(type))
 				continue;
 
@@ -252,22 +294,64 @@ public class Stub {
 		return type.FullName.Substring(start);
 	}
 	
-	public static void DoUpdateAssembly(string source, string dest) {
+	private static void AddIndexAssembly (Assembly assembly, XmlElement parent)
+	{
+		XmlElement index_assembly = parent.OwnerDocument.CreateElement("Assembly");
+		index_assembly.SetAttribute("Name", assembly.GetName().Name);
+		index_assembly.SetAttribute("Version", assembly.GetName().Version.ToString());
+		MakeAttributes(index_assembly, assembly, true);
+		parent.AppendChild(index_assembly);
+	}
+
+	private static void DoUpdateAssemblies (string source, string dest) 
+	{
 		string indexfile = dest + "/index.xml";
 		XmlDocument index;
 		if (new System.IO.FileInfo(indexfile).Exists) {
 			index = new XmlDocument();
 			index.Load(indexfile);
+
+			// Format change
+			ClearElement(index.DocumentElement, "Assembly");
+			ClearElement(index.DocumentElement, "Attributes");
 		} else {
 			index = CreateIndexStub();
 		}
-		XmlElement index_types = (XmlElement)index.SelectSingleNode("Overview/Types");
 		
-		MakeAttributes(index.DocumentElement, assembly, true);
+		if (name == null) {
+			string defaultTitle = "Untitled";
+			if (assemblies.Length == 1)
+				defaultTitle = assemblies[0].GetName().Name;
+			WriteElementInitialText(index.DocumentElement, "Title", defaultTitle);
+		} else {
+			WriteElementText(index.DocumentElement, "Title", name);
+		}
+		
+		XmlElement index_types = WriteElement(index.DocumentElement, "Types");
+		XmlElement index_assemblies = WriteElement(index.DocumentElement, "Assemblies");
+		index_assemblies.RemoveAll ();
 		
 		Hashtable goodfiles = new Hashtable();
+
+		foreach (Assembly assm in assemblies) {
+			AddIndexAssembly (assm, index_assemblies);
+			DoUpdateAssembly (assm, index_types, source, dest, goodfiles);
+		}
 		
+		CleanupFiles (dest, goodfiles);
+		CleanupIndexTypes (index_types, goodfiles);
+
+		using (System.IO.TextWriter writer
+			= new System.IO.StreamWriter(new System.IO.FileStream(indexfile, System.IO.FileMode.Create)))
+			WriteXml(index.DocumentElement, writer);
+	
+	}
+		
+	private static void DoUpdateAssembly (Assembly assembly, XmlElement index_types, string source, string dest, Hashtable goodfiles) 
+	{
 		foreach (Type type in assembly.GetTypes()) {
+			if (type.Namespace == null) continue;
+			
 			// Must get the A+B form of the type name.
 			string typename = GetTypeFileName(type);
 			
@@ -276,7 +360,6 @@ public class Stub {
 			if (file.Exists) {
 				// Update
 				XmlDocument basefile = new XmlDocument();
-				basefile.PreserveWhitespace = true;
 				try {
 					basefile.Load(typefile);
 				} catch (Exception e) {
@@ -306,17 +389,17 @@ public class Stub {
 			// Add namespace and type nodes into the index file as needed
 			XmlElement nsnode = (XmlElement)index_types.SelectSingleNode("Namespace[@Name='" + type.Namespace + "']");
 			if (nsnode == null) {
-				nsnode = index.CreateElement("Namespace");
+				nsnode = index_types.OwnerDocument.CreateElement("Namespace");
 				nsnode.SetAttribute("Name", type.Namespace);
 				index_types.AppendChild(nsnode);
 			}
 			XmlElement typenode = (XmlElement)nsnode.SelectSingleNode("Type[@Name='" + typename + "']");
 			if (typenode == null) {
-				typenode = index.CreateElement("Type");
+				typenode = index_types.OwnerDocument.CreateElement("Type");
 				typenode.SetAttribute("Name", typename);
 				nsnode.AppendChild(typenode);
 			}
-			
+				
 			// Ensure the namespace index file exists
 			if (!new System.IO.FileInfo(dest + "/" + type.Namespace + ".xml").Exists) {
 				Console.Error.WriteLine("New Namespace File: " + type.Namespace);
@@ -326,7 +409,10 @@ public class Stub {
 			// mark the file as corresponding to a type
 			goodfiles[type.Namespace + "/" + typename + ".xml"] = goodfiles;
 		}
-		
+	}
+
+	private static void CleanupFiles (string dest, Hashtable goodfiles)
+	{
 		// Look for files that no longer correspond to types
 		foreach (System.IO.DirectoryInfo nsdir in new System.IO.DirectoryInfo(dest).GetDirectories("*")) {
 			foreach (System.IO.FileInfo typefile in nsdir.GetFiles("*.xml")) {
@@ -338,19 +424,18 @@ public class Stub {
 				}
 			}
 		}
+	}
 		
+	private static void CleanupIndexTypes (XmlElement index_types, Hashtable goodfiles)
+	{
 		// Look for type nodes that no longer correspond to types
 		foreach (XmlElement typenode in index_types.SelectNodes("Namespace/Type")) {
 			string fulltypename = ((XmlElement)typenode.ParentNode).GetAttribute("Name") + "/" + typenode.GetAttribute("Name") + ".xml";
 			if (!goodfiles.ContainsKey(fulltypename))
 				typenode.ParentNode.RemoveChild(typenode);
 		}
-		
-		using (System.IO.TextWriter writer
-			= new System.IO.StreamWriter(new System.IO.FileStream(indexfile, System.IO.FileMode.Create)))
-			WriteXml(index.DocumentElement, writer);
 	}
-	
+		
 	public static void DoUpdateType2(XmlDocument basefile, Type type, string output) {
 		Console.Error.WriteLine("Updating: " + type.FullName);
 		
@@ -409,12 +494,26 @@ public class Stub {
 				XmlElement mm = MakeMember(basefile, m);
 				if (mm == null) continue;
 				members.AppendChild( mm );
-				members.AppendChild( members.OwnerDocument.CreateWhitespace("\n  ") );
 	
 				Console.Error.WriteLine("Member Added: " + mm.SelectSingleNode("MemberSignature/@Value").InnerText);
 				additions++;
 			}
 		}
+		
+		// Import code snippets from files
+		foreach (XmlNode code in basefile.GetElementsByTagName("code")) {
+			if (!(code is XmlElement)) continue;
+			string file = ((XmlElement)code).GetAttribute("src");
+			if (file != "") {
+				try {
+					using (System.IO.StreamReader reader = new System.IO.StreamReader(srcPath + "/" + file)) {
+						code.InnerText = reader.ReadToEnd();
+					}
+				} catch (Exception e) {
+					Console.Error.WriteLine("Could not load code file '" + srcPath + "/" + file + "': " + e.Message);
+				}
+			}
+		}
 
 		System.IO.TextWriter writer;
 		if (output == null)
@@ -516,10 +615,6 @@ public class Stub {
 	
 	// CREATE A STUB DOCUMENTATION FILE	
 
-	public static XmlElement StubType(string type) {
-		return StubType(assembly.GetType(type, true));
-	}
-	
 	public static XmlElement StubType(Type type) {
 		string typesig = MakeTypeSignature(type);
 		if (typesig == null) return null; // not publicly visible
@@ -629,14 +724,31 @@ public class Stub {
 		if (mi is FieldInfo && GetFieldConstValue((FieldInfo)mi, out fieldValue))
 			WriteElementText(me, "MemberValue", fieldValue);
 		
-		if (mi is MethodInfo)
-			MakeDocNode(me, ((MethodInfo)mi).GetParameters(), ((MethodInfo)mi).ReturnType, true, true);
-		else if (mi is ConstructorInfo)
-			MakeDocNode(me, ((ConstructorInfo)mi).GetParameters(), null, false, true);
+		ParameterInfo[] paraminfo = null;
+		Type returntype = null;
+		bool returnisreturn = true; // the name of the node is 'return', vs 'value'
+		bool addremarks = true;
+		
+		if (mi is MethodInfo || mi is ConstructorInfo)
+			paraminfo = ((MethodBase)mi).GetParameters();
 		else if (mi is PropertyInfo)
-			MakeDocNode(me, ((PropertyInfo)mi).GetIndexParameters(), ((PropertyInfo)mi).PropertyType, false, true);
-		else
-			MakeDocNode(me, (ParameterInfo[])null, null, false, mi.DeclaringType.IsEnum ? false : true);
+			paraminfo = ((PropertyInfo)mi).GetIndexParameters();
+			
+		if (mi is MethodInfo) {
+			returntype = ((MethodInfo)mi).ReturnType;
+		} else if (mi is PropertyInfo) {
+			returntype = ((PropertyInfo)mi).PropertyType;
+			returnisreturn = false;
+		}
+
+		// no remarks section for enum members
+		if (mi.DeclaringType.IsEnum)
+			addremarks = false;
+
+		string slashdocsig = MakeSlashDocMemberSig(mi);
+
+		MakeDocNode(me, paraminfo, returntype, returnisreturn, addremarks, slashdocsig);
+		
 	}
 	
 	private static bool GetFieldConstValue(FieldInfo field, out string value) {
@@ -708,15 +820,15 @@ public class Stub {
 			parameters = type.GetMethod("Invoke").GetParameters();
 			returntype = type.GetMethod("Invoke").ReturnType;
 		}
-		MakeDocNode(parameters, returntype, true, e, addremarks);
+		MakeDocNode(parameters, returntype, true, e, addremarks, "T:" + type.FullName);
 	}
 	
-	private static void MakeDocNode(XmlElement root, ParameterInfo[] parameters, Type returntype, bool returnisreturn, bool addremarks) {
+	private static void MakeDocNode(XmlElement root, ParameterInfo[] parameters, Type returntype, bool returnisreturn, bool addremarks, string slashdocsig) {
 		XmlElement e = WriteElement(root, "Docs");
-		MakeDocNode(parameters, returntype, returnisreturn, e, addremarks);
+		MakeDocNode(parameters, returntype, returnisreturn, e, addremarks, slashdocsig);
 	}
 	
-	private static void MakeDocNode(ParameterInfo[] parameters, Type returntype, bool returnisreturn, XmlElement e, bool addremarks) {
+	private static void MakeDocNode(ParameterInfo[] parameters, Type returntype, bool returnisreturn, XmlElement e, bool addremarks, string slashdocname) {
 		WriteElementInitialText(e, "summary", "To be added.");
 		
 		if (parameters != null) {
@@ -732,13 +844,13 @@ public class Stub {
 			if (parameters.Length == 1) {
 				int c = e.SelectNodes("param").Count;
 				if (c == 1) {
-					((XmlElement)e.SelectSingleNode("param")).SetAttribute("name", parameters[0].Name);
+					UpdateParameterName (e, (XmlElement) e.SelectSingleNode("param"), parameters[0].Name);
 				} else if (c > 1) {
 					foreach (XmlElement paramnode in e.SelectNodes("param")) {
 						if (paramnode.InnerText.StartsWith("To be added"))
 							paramnode.ParentNode.RemoveChild(paramnode);
 						else
-							paramnode.SetAttribute("name", parameters[0].Name);
+							UpdateParameterName (e, paramnode, parameters[0].Name);
 					}
 				}
 			}
@@ -798,8 +910,9 @@ public class Stub {
 			}
 		}
 		
+		string retnodename = null;
 		if (returntype != null && returntype != typeof(void)) {
-			string retnodename = returnisreturn ? "returns" : "value";
+			retnodename = returnisreturn ? "returns" : "value";
 			string retnodename_other = !returnisreturn ? "returns" : "value";
 			
 			// If it has a returns node instead of a value node, change its name.
@@ -820,9 +933,47 @@ public class Stub {
 		if (addremarks)
 			WriteElementInitialText(e, "remarks", "To be added.");
 		
+		if (slashdocs != null && slashdocname != null) {
+			XmlElement elem = (XmlElement)slashdocs.SelectSingleNode("doc/members/member[@name='" + slashdocname + "']");
+			if (elem != null) {
+				XmlElement xsummary = (XmlElement)elem.SelectSingleNode("summary");
+				if (xsummary != null)
+					WriteElementText(e, "summary", xsummary.InnerText);
+
+				XmlElement xremarks = (XmlElement)elem.SelectSingleNode("remarks");
+				if (xremarks != null)
+					WriteElementText(e, "remarks", xremarks.InnerText);
+				else if (xsummary != null)
+					ClearElement(e, "remarks");
+
+				if (retnodename != null) {
+					XmlElement xreturns = (XmlElement)elem.SelectSingleNode(retnodename);
+					if (xreturns != null)
+						WriteElementText(e, retnodename, xreturns.InnerText);
+				}
+				
+				foreach (XmlElement p in elem.SelectNodes("param")) {
+					XmlElement p2 = (XmlElement)e.SelectSingleNode("param[@name='" + p.GetAttribute("name") + "']");
+					if (p2 != null)
+						p2.InnerText = p.InnerText;
+				}
+			}
+		}
+		
 		NormalizeWhitespace(e);
 	}
-	
+
+	private static void UpdateParameterName (XmlElement docs, XmlElement pe, string newName)
+	{
+		string existingName = pe.GetAttribute ("name");
+		pe.SetAttribute ("name", newName);
+		if (existingName == newName)
+			return;
+		foreach (XmlElement paramref in docs.SelectNodes (".//paramref"))
+			if (paramref.GetAttribute ("name").Trim () == existingName)
+				paramref.SetAttribute ("name", newName);
+	}
+
 	private static void NormalizeWhitespace(XmlElement e) {
 		// Remove all text and whitespace nodes from the element so it
 		// is outputted with nice indentation and no blank lines.
@@ -1111,7 +1262,6 @@ public class Stub {
 		StringBuilder sb = new StringBuilder ();
 
 		int i = 0;
-		string modifier;
 		foreach (ParameterInfo parameter in pi) {
 			if (i != 0) sb.Append (", ");
 			if (parameter.ParameterType.IsByRef) {
@@ -1175,15 +1325,19 @@ public class Stub {
 
 
 	static string MakePropertySignature (PropertyInfo property) {
-		// pick an accessor
+		// Check accessibility of get and set, since they can be different these days.
+		string get_visible = null, set_visible = null;
+		MethodBase get_method = property.GetGetMethod (true);
+		MethodBase set_method = property.GetSetMethod (true);
+		if (get_method != null) get_visible = GetMethodVisibility(get_method);
+		if (set_method != null) set_visible = GetMethodVisibility(set_method);
+		if (get_visible == null && set_visible == null) return null; // neither are visible
+		
+		// Pick an accessor to use for static/virtual/override/etc. checks.
 		MethodBase method = property.GetSetMethod (true);
 		if (method == null)
 			method = property.GetGetMethod (true);
 	
-		string visibility = GetMethodVisibility(method);
-		if (visibility == null)
-			return null;
-	
 		string modifiers = String.Empty;
 		if (method.IsStatic) modifiers += " static";
 		if (method.IsVirtual && !method.IsAbstract) {
@@ -1202,13 +1356,26 @@ public class Stub {
 		string parameters = GetMethodParameters (property.GetIndexParameters());
 		if (parameters != "") parameters = "[" + parameters + "]";		
 		
-		string accessors = null;
-		if (property.CanRead && property.CanWrite)
-			accessors = "{ set; get; }";
-		else if (property.CanRead)
-			accessors = "{ get; }";
-		else if (property.CanWrite)
-			accessors = "{ set; }";
+		string visibility;
+		if (get_visible != null && (set_visible == null || (set_visible != null && get_visible == set_visible)))
+			visibility = get_visible;
+		else if (set_visible != null && get_visible == null)
+			visibility = set_visible;
+		else
+			visibility = "public"; // if they are different, but both externally accessible, the greater access must be public, I think
+		
+		string accessors = "{";
+		if (set_visible != null) {
+			if (set_visible != visibility)
+				accessors += " " + set_visible;
+			accessors += " set;";
+		}
+		if (get_visible != null) {
+			if (get_visible != visibility)
+				accessors += " " + get_visible;
+			accessors += " get;";
+		}
+		accessors += " }";
 	
 		return String.Format ("{0}{1} {2} {3}{4} {5};",
 						visibility, modifiers, type_name, name, parameters, accessors);
@@ -1301,5 +1468,39 @@ public class Stub {
 		return type;
 	}
 	
+	private static string MakeSlashDocMemberSig(MemberInfo m) {
+		string typename = m.DeclaringType.FullName;
+		string memname = m.Name;
+		
+		if (m is ConstructorInfo)
+			return "C:" + typename + ".#ctor(" + MakeSlashDocMemberSigArgs(((ConstructorInfo)m).GetParameters()) + ")";
+		if (m is MethodInfo)
+			return "M:" + typename + "." + memname + MakeSlashDocMemberSigArgs(((MethodInfo)m).GetParameters()) + MakeSlashDocMemberSigReturn((MethodInfo)m);
+		if (m is PropertyInfo && ((PropertyInfo)m).GetIndexParameters().Length == 0)
+			return "P:" + typename + "." + memname;
+		if (m is PropertyInfo && ((PropertyInfo)m).GetIndexParameters().Length > 0)
+			return "P:" + typename + "." + memname + MakeSlashDocMemberSigArgs(((PropertyInfo)m).GetIndexParameters());
+		if (m is FieldInfo)
+			return "F:" + typename + "." + memname;
+		if (m is EventInfo)
+			return "F:" + typename + "." + memname;
+		throw new ArgumentException();
+	}
+	
+	private static string MakeSlashDocMemberSigReturn(MethodInfo m) {
+		if (m.Name != "op_Implicit" && m.Name != "op_Explicit") return "";
+		return "~" + m.ReturnType.FullName;
+	}
+	
+	private static string MakeSlashDocMemberSigArgs(ParameterInfo[] ps) {
+		string ret = "";
+		foreach (ParameterInfo p in ps) {
+			if (ret != "") ret += ",";
+			ret += p.ParameterType.FullName.Replace("&", "@");
+		}
+		if (ret != "")
+			ret = "(" + ret + ")";
+		return ret;
+	}
 
 }
